import asyncio
import threading
import time
import logging

from xknx import XKNX
from xknx.io import ConnectionConfig, ConnectionType
from xknx.core import XknxConnectionState
from xknx.telegram import Telegram
from xknx.tools import (
    group_value_read,
    group_value_response,
    group_value_write,
    read_group_value
)

# from xknx.telegram.address import DeviceGroupAddress
from xknx.telegram.apci import GroupValueRead, GroupValueResponse, GroupValueWrite
from xknx.io import GatewayScanner
# from xknx.devices import Switch

# https://bbc.github.io/cloudfit-public-docs/asyncio/asyncio-part-5.html

class KNXValue(object):
    def __init__(self, parent, address, timeout=None, data=None):
        self._parent=parent
        self._address=address
        self._data=data
        self._timeout=timeout
        self._stamp=0

        logging.getLogger("xknx.log").disabled=True
        logging.getLogger("xknx.knx").disabled=True

    @property
    def parent(self):
        return self._parent

    @property
    def client(self):
        try:
            return self._parent._client
        except:
            pass

    @property
    def address(self):
        return self._address

    @property
    def data(self):
        return self._data

    def setData(self, data):
        if data is not None:
            self._data=data
            self._stamp=time.time()

    def getData(self):
        return self._data

    def age(self):
        return time.time()-self._stamp

    def isTimeout(self):
        if self._timeout:
            if self.age()>=self._timeout:
                return True
        return False

    def refresh(self):
        try:
            self.client.refreshgroup(self.address)
        except:
            pass

    def refreshIfTimeout(self):
        if self.isTimeout():
            self.refresh()

    def write(self, data=None):
        try:
            if data is None:
                data=self.getData()
            if data is not None:
                self.parent.client.writegroup(self.address, data)
                self.refresh()
        except:
            pass

    def rewrite(self):
        self.write(self.data)


class KNXValues(object):
    def __init__(self, client, timeout=15):
        self._client=client
        self._timeout=timeout
        self._values={}

    @property
    def client(self):
        return self._client

    def get(self, address) -> KNXValue:
        try:
            return self._values[address]
        except:
            pass

    def add(self, address, timeout=None):
        if address:
            value=self.get(address)
            if value is None:
                if timeout is None:
                    timeout=self._timeout

                value=KNXValue(self, address, timeout=timeout)
                self._values[address]=value
            return value

    def set(self, address, data):
        value=self.add(address)
        if value is not None:
            value.setData(data)

    def update(self, address, data):
        value=self.get(address)
        if value is not None:
            value.setData(data)

    def getData(self, address):
        value=self.get(address)
        if value is not None:
            return value.getData()

    def refresh(self):
        for value in self._values.values():
            value.refresh()

    def refreshIfTimeout(self):
        for value in self._values.values():
            if value.isTimeout():
                value.refresh()


class XKNXClient(object):
    def __init__(self, logger, host=None, timeoutRefresh=15):
        self._logger=logger
        self._host=host
        self._xknx=None
        self._connectState=None
        self._loop=None
        self._thread=None
        self._values=KNXValues(self, timeout=timeoutRefresh)

    @property
    def logger(self):
        return self._logger

    def setHost(self, host):
        if host:
            self._host=host

    def getValue(self, address) -> KNXValue:
        return self._values.get(address)

    def getValueData(self, address) -> KNXValue:
        value=self._values.get(address)
        if value is not None:
            return value.data

    def setValue(self, address, data):
        if data:
            value=self._values.add(address)
            if value is not None:
                value.setData(data)

    def refreshValues(self):
        self._values.refresh()

    def refreshValuesIfTimeout(self):
        self._values.refreshIfTimeout()

    def onConnectionStateChanged(self, state: XknxConnectionState):
        self.logger.debug(state)
        self._connectState=state

    def onGroupValueResponse(self, telegram: Telegram):
        self.logger.debug('KNX: %s' % telegram)
        address=str(telegram.destination_address)
        data=telegram.payload.value.value
        self._values.update(address, data)

    def onGroupValueRead(self, telegram: Telegram):
        # as said by xknx, usually not trapped
        self.logger.debug('KNX: %s' % telegram)
        pass

    def onGroupValueWrite(self, telegram: Telegram):
        # as said by xknx, usually not trapped
        self.logger.debug('KNX: %s' % telegram)
        pass

    def onTelegramReceived(self, telegram: Telegram):
        if isinstance(telegram.payload, GroupValueWrite):
            self.onGroupValueWrite(telegram)
        elif isinstance(telegram.payload, GroupValueResponse):
            self.onGroupValueResponse(telegram)
        elif isinstance(telegram.payload, GroupValueRead):
            self.onGroupValueRead(telegram)
        else:
            self.logger.warning('KNX: unhandled telegram %s' % telegram)

    async def connect(self):
        if not self._xknx and self._host:
            config=ConnectionConfig(
                connection_type=ConnectionType.TUNNELING,
                gateway_ip=self._host)

            self.logger.info('KNX: connecting to gateway %s...' % self._host)
            self._xknx=XKNX(connection_config=config,
                    daemon_mode=True,
                    state_updater=True,
                    connection_state_changed_cb=self.onConnectionStateChanged,
                    telegram_received_cb=self.onTelegramReceived)

        if self._xknx:
            await self._xknx.start()

    async def disconnect(self):
        if self._xknx:
            self.logger.info('KNX: disconnecting from gateway %s...' % self._host)
            await self._xknx.stop()
            self._xknx=None

        if self._loop:
            self._loop.stop()

    def asyncioRunOnce(self):
        self.submit(asyncio.sleep(0))

    def submit(self, coro, timeout=0):
        if self.isRunning():
            try:
                future=asyncio.run_coroutine_threadsafe(coro, self._loop)
                result=future.result(timeout)
                return result
            except TimeoutError:
                pass
            except:
                # FIXME:
                self.logger.exception('KNX: submit()')
                pass

    def submitNoWait(self, coro):
        if self.isRunning():
            try:
                asyncio.run_coroutine_threadsafe(coro, self._loop)
            except:
                pass

    # def switch(self, name, addres):
        # device=Switch(self._xknx, name, addres)
        # self._xknx.devices.async_add(device)

    def readgroup(self, address, timeout=2.0):
        if self.isRunning():
            return self.submit(read_group_value(self._xknx, address), timeout)

    def refreshgroup(self, address):
        if self.isRunning():
            # response will come later via callback
            self.submitNoWait(read_group_value(self._xknx, address))

    def writegroup(self, address, value, value_type=None):
        # FIXME: why this call return None (not as read) ?
        # as coro is None, we can't sumbit it to the asyncio loop
        coro=group_value_write(self._xknx, address, value, value_type)
        if coro is not None:
            self.logger.warning('KNX: coroutine is not None, which seems good but not handled')

        # FIXME: without this, the telegram is not sent until next message
        self.asyncioRunOnce()

    def knxmanager(self):
        try:
            asyncio.set_event_loop(self._loop)
            asyncio.run_coroutine_threadsafe(self.connect(), self._loop)
            self.logger.info('KNX: async loop started')
            self._loop.run_forever()
            self.logger.info('KNX: async loop halted')
        except KeyboardInterrupt:
            self.logger.info("KNX: STOP!")
        except:
            pass

    def start(self):
        if not self._thread:
            self._loop = asyncio.new_event_loop()
            self._thread=threading.Thread(target=self.knxmanager)
            self._thread.daemon=True
            self._thread.start()
            self.logger.info('KNX: thread started')

    def stop(self):
        try:
            if self._loop and self._loop.is_running():
                asyncio.run_coroutine_threadsafe(self.disconnect(), self._loop)
        except:
            if self._loop:
                self.logger.info('KNX: stopping aysyncio loop')
                self._loop.stop()

        if self._thread:
            # wait for thread (loop) termination
            self._thread.join()
            self.logger.info('KNX: thread halted')
            if self._loop:
                self.logger.info('KNX: closing asyncio loop')
                self._loop.close()
                self._loop=None
            self._thread=None
            self._connectState=None

    def isRunning(self):
        if self._thread:
            return True
        return False

    def isConnected(self):
        if self.isRunning():
            if self._connectState==XknxConnectionState.CONNECTED:
                return True
        return False

    def discover(self, interface, stopOnFound=False):
        gateways={}

        scanner=GatewayScanner(self._xknx, local_ip=interface, stop_on_found=stopOnFound)
        items=self.submit(scanner.scan(), 15)

        try:
            for gateway in items:
                self.logger.debug(gateway)
                data={'address': gateway.individual_address,
                    'name': gateway.name.strip(),
                    'host': gateway.ip_addr, 'port': gateway.port,
                    'secure': gateway.supports_secure,
                    'tcp': gateway.supports_tunnelling_tcp,
                    'udp': gateway.supports_tunnelling}

                gateways[gateway.ip_addr]=data
        except:
            pass

        return gateways

    def __del__(self):
        self.stop()


if __name__ == '__main__':
    pass



from .task import MBIOTask
from .xmlconfig import XMLConfig

import math


class MBIOTaskVirtualIO(MBIOTask):
    def initName(self):
        return 'vio'

    def onInit(self):
        pass

    def compileExpression(self, expression):
        try:
            code=compile(expression, '<string>', 'eval')
            if not code:
                self.logger.error('Unable to compile expression [%s]' % expression)

            variables={}
            for name in code.co_names:
                variables[name]=None

            return (code, variables)
        except:
            pass

    def evalExpression(self, code, variables={}):
        try:
            if code:
                r=eval(code, None, variables)
                return r
        except:
            pass
        return None

    def onLoad(self, xml: XMLConfig):
        items=xml.children('digital')
        if items:
            for item in items:
                name=item.get('name')
                if name:
                    self.valueDigital(name, default=item.getBool('default'), writable=True)

        items=xml.children('analog')
        if items:
            for item in items:
                name=item.get('name')
                unit=item.get('unit')
                resolution=item.getFloat('resolution', 0.1)
                if name:
                    self.value(name, unit=unit, default=item.getFloat('default'), writable=True, resolution=resolution)

        items=xml.children('variable')
        if items:
            for item in items:
                name=item.get('name')
                unit=item.get('unit')
                resolution=item.getFloat('resolution', 0.1)
                if name:
                    value=self.value(name, unit=unit, default=item.getFloat('default'), resolution=resolution)
                    value.config.set('x0', 0.0)
                    value.config.xmlUpdateFloat(item, 'x0', vmin=0)
                    value.config.set('x1', 100.0)
                    value.config.xmlUpdateFloat(item, 'x1', vmin=value.config.x0)
                    value.config.set('y0', 0.0)
                    value.config.xmlUpdateFloat(item, 'y0')
                    value.config.set('y1', 100.0)
                    value.config.xmlUpdateFloat(item, 'y1', vmin=value.config.y0)
                    expression=item.get('expression')
                    if expression:
                        value.config.set('expression', expression)
                        code, variables=self.compileExpression(expression)
                        value.config.set('code', code)
                        value.config.set('variables', variables)

    def poweron(self):
        return True

    def poweroff(self):
        return True

    def eval(self, value):
        if value is not None:
            code=value.config.code
            if code is not None:
                variables=value.config.variables
                if variables:
                    mbio=self.getMBIO()
                    for name in variables.keys():
                        try:
                            variables[name]=mbio.value(name).value
                        except:
                            pass

                # self.logger.warning('%s <- %s' % (value, variables))

                v=None
                r=self.evalExpression(code, variables)
                if r is not None:
                    value.setError(False)
                    v=r

                try:
                    dy=(value.config.y1-value.config.y0)
                    dx=(value.config.x1-value.config.x0)
                    v=value.config.y0+(v-value.config.x0)/dx*dy
                    if v<value.config.y0:
                        v=value.config.y0
                    if v>value.config.y1:
                        v=value.config.y1
                except:
                    pass

                value.updateValue(v)
            else:
                value.setError(True)

    def run(self):
        for value in self.values:
            self.microsleep()

            if value.config.expression:
                self.eval(value)

            if value.isWritable() and value.isPendingSync():
                value.clearSyncAndUpdateValue()
        return 1.0


if __name__ == "__main__":
    pass



import functools

from http.server import ThreadingHTTPServer, SimpleHTTPRequestHandler
import threading
import time
import shutil
import pathlib
import requests

# warning: importing pathlib replace any "import os" order

# https://nachtimwald.com/2019/12/10/python-http-server/
# SSL: https://realpython.com/python-http-server/


class QuietSimpleHTTPRequestHandler(SimpleHTTPRequestHandler):
    def log_message(self, format, *args):
        pass

    def log_error(self, *args):
        pass


class TemporaryWebServer(object):
    def __init__(self, fpath='/tmp/wserver', port=8000, host=None):
        self._path: pathlib.Path = pathlib.Path(fpath)
        self._host=host
        self._port=int(port)
        self._interface=''
        self._thread=None
        self._httpd=None
        self._files={}
        self._timeout=0

    def url(self, path=None):
        host=self._host or self._interface
        if host:
            url='http://%s:%d' % (host, self._port)
            if path:
                if path[0]!='/':
                    url += '/'
                url += path
            return url

    def server(self):
        handler=functools.partial(QuietSimpleHTTPRequestHandler, directory=self.pathstr())
        with ThreadingHTTPServer((self._interface, self._port), handler) as httpd:
            try:
                self._httpd=httpd
                httpd.serve_forever()
            except:
                pass
            httpd.server_close()

        self._httpd=None

    def pathstr(self):
        try:
            return str(self._path)
        except:
            pass

    def createPath(self):
        try:
            self._path.mkdir(parents=True, exist_ok=True)
            return True
        except:
            pass
        return False

    def getPathForFile(self, fname):
        try:
            return self._path.joinpath(fname)
        except:
            pass

    def isFile(self, fname):
        try:
            p=self.getPathForFile(fname)
            if p.exists() and p.is_file():
                return True
        except:
            pass
        return False

    def importFile(self, fpath, fname=None, timeout=0):
        try:
            sp=pathlib.Path(fpath)
            if sp.is_file():
                self.start()
                if not fname:
                    fname=sp.name
                tp=self.getPathForFile(fname)
                shutil.copyfile(str(sp), str(tp))
                self._files[fname]={'fname': fname, 'timeout': timeout}
                self._timeout=time.time()+60
                url=self.url(fname)
                if url:
                    return url
        except:
            pass
        return False

    def getFileContent(self, fname):
        try:
            if fname:
                p=self.getPathForFile(fname)
                with open(str(p), 'rb') as f:
                    data=f.read()
                    return data
        except:
            pass

    def removeFile(self, fname):
        try:
            if fname:
                p=self.getPathForFile(fname)
                p.unlink()
                try:
                    del self._files[fname]
                except:
                    pass
                return True
        except:
            pass
        return False

    def isFileTimeout(self, fname):
        try:
            timeout=self._files[fname]['timeout']
            if timeout>0 and self.isFile(fname):
                p=self.getPathForFile(fname)
                info=p.stat()
                age=time.time()-info.st_mtime
                if age>=timeout:
                    return True
        except:
            pass
        return False

    def getFiles(self):
        try:
            files=[]
            for f in self._path.iterdir():
                fname=f.name
                if not self.isFile(fname):
                    continue
                if self.isFileTimeout(fname):
                    self.removeFile(fname)
                    continue
                files.append(fname)
            return files
        except:
            pass

    def start(self):
        if not self._thread:
            self.createPath()
            self._thread=threading.Thread(target=self.server)
            self._thread.daemon=True
            self._thread.start()

    def stop(self):
        if self._thread:
            try:
                self._httpd.shutdown()
            except:
                pass

            try:
                # force a fake request to the server (may be waiting for a request)
                requests.get('http://localhost:%d/fake' % self._port, timeout=1.0)
            except:
                pass

            self._thread.join()
            self._thread=None

    def isRunning(self):
        if self._thread:
            return True
        return False

    def manager(self):
        files=self.getFiles()
        if files:
            self._timeout=time.time()+60

        if time.time()>self._timeout:
            self.stop()

    def __del__(self):
        self.stop()


if __name__ == '__main__':
    ws=TemporaryWebServer('/tmp/wserver')
    ws.importFile('/tmp/myfile', 'fhe', 600)
    ws.start()



#!/bin/python

from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .value import MBIOValue

from .task import MBIOTask
from .xmlconfig import XMLConfig


import requests

from urllib3.exceptions import InsecureRequestWarning
requests.packages.urllib3.disable_warnings(category=InsecureRequestWarning)

# https://www.sensorpush.com/gateway-cloud-api
# https://dashboard.sensorpush.com/


class MBIOTaskSensorPush(MBIOTask):
    def initName(self):
        return 'spush'

    def onInit(self):
        self.config.set('refreshperiod', 60)
        self._token=None
        self._timeoutToken=0
        self._timeoutRefresh=0
        self._retry=3
        self.valueDigital('comerr', default=False)
        self._cache={'gateways': None, 'sensors': None}
        self._data=[]
        self._sensors=[]

    def onLoad(self, xml: XMLConfig):
        self.config.set('email', xml.get('email'))
        self.config.set('password', xml.get('password'))
        self.config.update('refreshperiod', xml.getInt('refresh'))
        self._data=[]
        for item in xml.children('sensor'):
            sid=item.get('id')
            if sid:
                for v in item.children('value'):
                    name=v.get('name')
                    dataname=v.get('source')
                    if name and dataname:
                        unit=v.get('unit')
                        resolution=v.getFloat('resolution', 0.1)
                        default=v.getFloat('default')
                        value=self.value(name, unit=unit, default=default, resolution=resolution)
                        item={'name': name, 'sid': sid, 'dataname': dataname, 'value': value}
                        self._data.append(item)
                        if sid not in self._sensors:
                            self._sensors.append(sid)

    @property
    def token(self):
        return self._token

    def url(self, path='/'):
        url='https://api.sensorpush.com/api/v1'
        return '%s/%s' % (url, path)

    def POST(self, path, data=None):
        try:
            url=self.url(path)
            headers={}
            if self._token:
                headers={'Authorization': self._token}
            if data is None:
                data={}

            # self.logger.debug('%s(%s)' % (url, data))
            self.logger.debug('%s' % (url))
            r=requests.post(url,
                            headers=headers, json=data,
                            verify=False, timeout=5.0)
            if r and r.ok:
                # self.logger.warning(r.text)
                data=r.json()
                # self.logger.debug(data)
                return data
            self.logger.error(r.text)
        except:
            self.logger.exception('post')

    def auth(self, force=False):
        if self._token and not self.isTimeout(self._timeoutToken):
            return self._token

        if not self.isTimeout(self._timeoutToken) and not force:
            return

        self._token=None
        self._timeoutToken=self.timeout(15)

        try:
            data={'email': self.config.email, 'password': self.config.password}
            r=self.POST('oauth/authorize', data)
            if r:
                key=r['authorization']
                if key:
                    data={'authorization': key}
                r=self.POST('oauth/accesstoken', data)
                if r:
                    self._token=r['accesstoken']
                    self._timeoutToken=self.timeout(60*45)
                    return self._token
        except:
            self.logger.exception('auth')

    def deauth(self):
        self._token=None

    def cache(self, name):
        try:
            return self._cache[name]
        except:
            pass

    def cacheUpdate(self, name, data):
        try:
            if name and data:
                self._cache[name]=data
        except:
            pass
        return self.cache(name)

    def retrieveGateways(self, refresh=False):
        cache='gateways'
        data=self.cache(cache)
        if refresh or not data:
            data={'active': True, 'format': 'json'}
            r=self.POST('devices/gateways', data)
            return self.cacheUpdate(cache, r)
        return data

    def gateways(self, refresh=False):
        gateways=[]
        try:
            for gateway in self.retrieveGateways(refresh).values():
                gateways.append(gateway['id'])
        except:
            pass
        return gateways

    def retrieveSensors(self, refresh=False):
        cache='sensors'
        data=self.cache(cache)
        if refresh or not data:
            data={'active': True, 'format': 'json'}
            r=self.POST('devices/sensors', data)
            return self.cacheUpdate(cache, r)
        return data

    def sensors(self, refresh=False):
        sensors=[]
        try:
            for sensor in self.retrieveSensors(refresh).keys():
                sensors.append(sensor)
        except:
            pass
        return sensors

    def sensorType(self, sid):
        try:
            return self.retrieveSensors()[sid]['type'].lower()
        except:
            pass

    def f2c(self, value):
        try:
            value=float(value)
            return (value-32.0)*5.0/9.0
        except:
            pass

    def retrieveValues(self, sensors=None):
        if not sensors:
            sensors=self.sensors()
        if sensors:
            data={'active': True, 'bulk': False, 'format': 'json',
                'limit': 1, 'sensors': sensors}
            r=self.POST('samples', data)
            try:
                result={}
                for sensor in sensors:
                    try:
                        state=r['sensors'][sensor][0]
                        result[sensor]=state
                    except:
                        pass
                return result
            except:
                pass

    def poweron(self):
        self.auth()
        self.retrieveGateways()
        self.retrieveSensors()
        return True

    def poweroff(self):
        self.deauth()
        return True

    def run(self):
        if not self.auth():
            return 5.0

        if self.isTimeout(self._timeoutRefresh):
            self._timeoutRefresh=self.timeout(self.config.refreshperiod)
            error=False

            dataread=self.retrieveValues(self._sensors)
            try:
                for item in self._data:
                    value=item['value']
                    try:
                        sid=item['sid']
                        stype=self.sensorType(sid)
                        sensordata=dataread[sid]
                        dataname=item['dataname']
                        svalue=float(sensordata[dataname])

                        if dataname=='temperature':
                            svalue=self.f2c(svalue)

                        value.updateValue(svalue)
                        value.setError(False)
                    except:
                        value.setError(True)
                        error=True
            except:
                self.logger.exception('spush')
                error=True

            if not error:
                self._timeoutRefresh=self.timeout(self.config.refreshperiod)
                self._retry=3
                self.values.comerr.updateValue(False)
            else:
                self._timeoutRefresh=self.timeout(60)
                if self._retry>0:
                    self._retry-=1
                    if self._retry==0:
                        self.values.comerr.updateValue(True)

        return 5.0


if __name__ == "__main__":
    pass



#!/bin/python

import xml.etree.ElementTree as ET

import base64


class XMLConfig(object):
    # xml is supposed to be already lowerized
    def __init__(self, xml: ET):
        self._xml=xml

    @property
    def tag(self):
        return self._xml.tag

    def isConfig(self, name):
        if self.tag==name.lower():
            return True
        return False

    def getAttribute(self, name, default=None):
        try:
            data=self._xml.get(name.lower(), default)
            if data[:5]=='(b16)':
                data=data[5:].upper()
                return base64.b16decode(data).decode()
            return data
        except:
            pass
        return default

    def get(self, name, default=None):
        return self.getAttribute(name, default)

    def text(self):
        try:
            return self._xml.text.strip()
        except:
            pass

    def __getkey__(self, key):
        return self.get(key)

    def hasAttribute(self, name):
        if self.getAttribute(name) is not None:
            return True
        return False

    def hasChild(self, name):
        if self.child(name):
            return True
        return False

    def getBool(self, name, default=None):
        try:
            value=self.getAttribute(name, default).lower()
            if value=='1' or value=='yes' or value=='true':
                return True
            return False
        except:
            pass
        return default

    def match(self, name, value):
        try:
            if self.get(name)==value.lower():
                return True
        except:
            pass
        return False

    def getInt(self, name, default=None, vmin=None, vmax=None):
        try:
            value=int(self.getAttribute(name, default))
            if vmin is not None:
                value=max(value, int(vmin))
            if vmax is not None:
                value=min(value, int(vmax))
            return value
        except:
            pass
        return default

    def getFloat(self, name, default=None, vmin=None, vmax=None):
        try:
            value=float(self.getAttribute(name, default))
            if vmin is not None:
                value=max(value, float(vmin))
            if vmax is not None:
                value=min(value, float(vmax))
            return value
        except:
            pass
        return default

    def getB16Str(self, name, default=None):
        try:
            # base64.b16encode(datastr.encode())
            data=self.get(name, default).upper().encode()
            return base64.b16decode(data).decode()
        except:
            pass
        return default

    def children(self, name='*'):
        children=[]
        items=self._xml.findall(name.lower())
        if items:
            for item in items:
                children.append(XMLConfig(item))
        return children

    def child(self, name):
        item=self._xml.find(name.lower())
        if item is not None:
            return XMLConfig(item)

    def tostring(self):
        return ET.tostring(self._xml, encoding='utf8')

    def dump(self):
        print(self.tostring())


if __name__ == "__main__":
    pass



#!/bin/python

from .task import MBIOTask
from .xmlconfig import XMLConfig

from .netscan import MBIONetworkScanner
import requests


# Should use https://shelly-api-docs.shelly.cloud/gen2/General/RPCChannels
# https://shelly-api-docs.shelly.cloud/gen2/General/ComponentConcept
# https://shelly-api-docs.shelly.cloud/gen2/ComponentsAndServices/Shelly
# Sys.GetConfig
# Shelly.GetConfig
# Shelly.ListMethods
# Shelly.GetDeviceInfo
# Shelly.GetComponents
# Switch.GetConfig?id=0
# Switch.GetStatus?id=0
# Switch.Toggle?id=0
# Switch.Set?id=0&on=true


class ShellyDevice(object):
    def __init__(self, ip, logger):
        self._ip=ip
        self._logger=logger
        self._methods=None
        self.onInit()

    def onInit(self):
        pass

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, self._ip)

    @property
    def logger(self):
        return self._logger

    def url(self, command):
        return 'http://%s/rpc/%s' % (self._ip, command)

    def doGET(self, command, params=None):
        try:
            url=self.url(command)
            self.logger.debug('shelly:%s(%s)' % (url, params))
            r=requests.get(url, params=params, timeout=3.0, verify=False)
            if r and r.ok:
                data=r.json()
                if data:
                    return data
                return True
        except:
            # self.logger.exception('doGET(%s:%s)' % (url, params))
            pass

    def call(self, name, params=None):
        return self.doGET(name, params)

    def getSystemStatus(self):
        return self.call('Sys.GetStatus')

    def getSystemConfig(self):
        return self.call('Sys.GetConfig')

    def reboot(self):
        return self.call('Shelly.Reboot')

    @property
    def firmware(self):
        try:
            return self.getSystemConfig()['device']['fw_id']
        except:
            pass

    @property
    def components(self):
        try:
            return self.call('Shelly.GetComponents')
        except:
            pass

    def upgrade(self, beta=False, url=None):
        stage='stable'
        if beta:
            stage='beta'
        params={'stage': stage}
        if url:
            params['url']=url
        return self.call('Shelly.Update', params)

    def getMAC(self):
        try:
            return self.getSystemConfig()['device']['mac'].lower()
        except:
            pass

    def ping(self):
        if self.getMAC():
            return True
        return False

    def methods(self, search=None):
        methods=self._methods
        if not methods:
            try:
                data=self.doGET('Shelly.ListMethods')
                methods=data['methods']
                if methods:
                    self._methods=methods
            except:
                pass
        if methods:
            if search:
                search=search.lower()
                return [item for item in methods if search in item.lower()]
            return methods


class ShellyDeviceSwitch(ShellyDevice):
    def getConfig(self, sid=0):
        params={'id': sid}
        return self.call('Switch.GetConfig', params)

    def getState(self, sid=0):
        try:
            params={'id': sid}
            data=self.call('Switch.GetStatus', params)
            if data['output']:
                return True
            return False
        except:
            pass

    def set(self, sid, state):
        try:
            if state:
                state='true'
            else:
                state='false'

            params={'id': sid, 'on': state}
            if self.call('Switch.Set', params):
                return True
        except:
            pass
        return False

    def on(self, sid=0):
        return self.set(sid, True)

    def off(self, sid=0):
        return self.set(sid, False)

    def toggle(self, sid=0):
        params={'id': sid}
        if self.call('Switch.Toggle', params):
            return True
        return False

class ShellyDeviceInput(ShellyDevice):
    def getConfig(self, sid=0):
        params={'id': sid}
        return self.call('Input.GetConfig', params)

    def getState(self, sid=0):
        try:
            params={'id': sid}
            data=self.call('Input.GetStatus', params)
            if data['state']:
                return True
            return False
        except:
            pass


class MBIOTaskShelly(MBIOTask):
    def initName(self):
        return 'shelly'

    def onInit(self):
        requests.packages.urllib3.disable_warnings()
        self.config.set('refreshperiod', 15)
        self._switches={}
        self._inputs={}
        self._devices=[]
        self._timeoutRefresh=0
        self.valueDigital('comerr', default=False)

    def addDevice(self, device):
        self._devices[device]

    def onLoad(self, xml: XMLConfig):
        self.config.update('refreshperiod', xml.getInt('refresh'))

        items=xml.children('switch')
        if items:
            for item in items:
                name=item.get('name')
                ip=item.get('ip')
                cid=item.getInt('id', 0)
                invert=item.getBool('invert', False)
                if name and not self._switches.get(name):
                    data={}
                    value=self.valueDigital('%s_state' % name, writable=True, commissionable=True)
                    data['state']=value
                    value.config.set('ip', ip)
                    value.config.set('cid', cid)
                    value.config.set('invert', invert)
                    value.config.set('device', self.switch(ip))
                    self._switches[name.lower()]=data

        items=xml.children('input')
        if items:
            for item in items:
                name=item.get('name')
                ip=item.get('ip')
                cid=item.getInt('id', 0)
                invert=item.getBool('invert', False)
                if name and not self._switches.get(name):
                    data={}
                    value=self.valueDigital('%s_state' % name, commissionable=True)
                    data['state']=value
                    value.config.set('ip', ip)
                    value.config.set('cid', cid)
                    value.config.set('invert', invert)
                    value.config.set('device', self.input(ip))
                    # data['t']=self.value('%s_t' % name, unit='C', resolution=0.1)
                    self._inputs[name.lower()]=data

    def poweron(self):
        return True

    def poweroff(self):
        return True

    def run(self):
        for name in self._switches.keys():
            value=self._switches[name]['state']
            if value.isPendingSync():
                self.microsleep()
                try:
                    device=value.config.device
                    cid=value.config.cid
                    if device.set(cid, value.toReachValue):
                        value.clearSync()
                        self._timeoutRefresh=0
                except:
                    pass
                value.clearSyncAndUpdateValue()

        if self.config.refreshperiod>0 and self.isTimeout(self._timeoutRefresh):
            self._timeoutRefresh=self.timeout(self.config.refreshperiod)
            error=False
            for name in self._switches.keys():
                self.microsleep()
                dev=self._switches[name]
                try:
                    value=dev['state']
                    device=value.config.device
                    cid=value.config.cid
                    v=device.getState(cid)
                    if v is not None:
                        if value.config.invert:
                            v=not v
                        value.updateValue(v)
                        value.setError(False)
                        continue
                    else:
                        error=True
                except:
                    pass

                dev['state'].setError(True)
                # dev['t'].setError(True)
                error=True

            for name in self._inputs.keys():
                self.microsleep()
                dev=self._inputs[name]
                try:
                    value=dev['state']
                    device=value.config.device
                    cid=value.config.cid
                    v=device.getState(cid)
                    if v is not None:
                        if value.config.invert:
                            v=not v
                        value.updateValue(v)
                        value.setError(False)
                        continue
                    else:
                        error=True
                except:
                    pass

                dev['state'].setError(True)
                # dev['t'].setError(True)
                error=True

            self.values.comerr.updateValue(error)

        return 5.0

    def scanner(self, network=None):
        network=network or self.getMBIO().network
        s=ShellyScanner(self, network)
        return s.scan()

    def device(self, ip):
        device=ShellyDevice(ip, self.logger)
        self._devices.append(device)
        return device

    def switch(self, ip):
        device=ShellyDeviceSwitch(ip, self.logger)
        self._devices.append(device)
        return device

    def input(self, ip):
        device=ShellyDeviceInput(ip, self.logger)
        self._devices.append(device)
        return device

    def upgrade(self):
        for device in self._devices:
            device.upgrade()

    def reboot(self):
        for device in self._devices:
            device.reboot()


class ShellyScanner(MBIONetworkScanner):
    def probe(self, host):
        try:
            url='http://%s/rpc/Sys.GetConfig' % (host)
            r=requests.get(url, timeout=1.0)
            if r and r.ok:
                data=r.json()
                return data['device']['mac']
        except:
            pass


if __name__ == "__main__":
    pass
